/* AGI 2.917 Interpreter Upgrader v1.0 by Cold Turkey, September 2017 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>

/* Byte arrays of updated interpreter files */
#include "files.h"


/* Example code from stack overflow. Allows searching inside
of buffers with several null terminated strings. Standard search
does not allow this                                          */
char *sstrstr(char *haystack, char *needle, int length)
{
    int needle_length = strlen(needle);
    int i;

    for (i = 0; i < length; i++)
    {
        if (i + needle_length > length)
        {
            return NULL;
        }

        if (strncmp(&haystack[i], needle, needle_length) == 0)
        {
            return &haystack[i];
        }
    }
    return NULL;
}

char* getInterpVer()
{
	char* searchString = "Adventure Game Interpreter";
	FILE* interp;
	//a healthy chunk of memory to hold the return val. Will definitely not need
	//all of this, but in 2014 this is peanuts!
	char* returnVal = (char *)malloc(256*sizeof(char));
	
	if ((interp = fopen("AGIDATA.OVL", "rb")) == NULL) 
	{
	  printf("Error opening AGIDATA.OVL, are you in a game directory?\n");
	  exit(1);
	}
	else
	{
		unsigned char    *buffer; // the memory to hold things
		long    numbytes; //the size of the file
		char* result;
		bool firstNumberFound;

		/* Get the number of bytes */
		fseek(interp, 0L, SEEK_END);
		numbytes = ftell(interp);

		/* reset the file position indicator to 
		the beginning of the file */
		fseek(interp, 0L, SEEK_SET);	

		/* grab sufficient memory for the 
		buffer to hold the text */
		if( (buffer = (unsigned char*)calloc(numbytes, sizeof(unsigned char))) == NULL)
		{
			printf("Can't allocate the memory to read AGIDATA.OVL \n");
			exit(1);
		}
		
		/* copy all the text into the buffer */
		fread(buffer, sizeof(unsigned char), numbytes, interp);
		fclose(interp);
		
		//If the string was found, loop to the end. The version is always null terminated!!
		
		if( (result = sstrstr((char*)buffer, searchString, numbytes)) != NULL)
		{
			//printf("%s", result);
			int i;
			int count = 0;
			firstNumberFound = false;
			for(i=0;i<strlen(result);i++)
			{
				if(isdigit(result[i]) && !firstNumberFound)
				{
					returnVal[count] = result[i];
					count++;
					firstNumberFound = true;
				}
				else if(firstNumberFound)
				{
					returnVal[count] = result[i];
					count++;
				}
			}
			//make sure we put the null at the end, or C will get confused :)
			returnVal[count] = 0;
		}
		else
		{
			printf("Interpretor version not found in original files, quitting... \n");
			exit(1);
		}
		
		/* free the memory we used for the buffer */
		free(buffer);
	}
	return returnVal;
}

bool removeGameFiles()
{
	system("mkdir backup");
	system("copy *.com backup");
	system("del *.com");
	system("copy AGI backup");
	system("copy CGA_GRAF.OVL backup");
	system("copy HGC_FONT backup");
	system("copy HGC_OBJS.OVL backup");
	system("copy JR_GRAF.OVL backup");
	system("copy VG_GRAF.OVL backup");
	system("copy AGIDATA.OVL backup");
	system("copy EGA_GRAF.OVL backup");
	system("copy HGC_GRAF.OVL backup");
	system("copy IBM_OBJS.OVL backup");
	
	if ((remove("AGI") == 0) && (remove("CGA_GRAF.OVL") == 0) && (remove("HGC_FONT") == 0) && (remove("HGC_OBJS.OVL") == 0) && (remove("JR_GRAF.OVL") == 0) && (remove("VG_GRAF.OVL") == 0) && (remove("AGIDATA.OVL") == 0) && (remove("EGA_GRAF.OVL") == 0) && (remove("HGC_GRAF.OVL") == 0) && (remove("IBM_OBJS.OVL") == 0))
	{
		return true;
	}
	else
	{
		return false;
	}
}

void writeNewGameFiles(char* id)
{
	FILE* file;
	char * gameCOM = (char *) malloc(strlen(id) + 8);
	strcpy(gameCOM, id);
	strcat(gameCOM, "new");
	strcat(gameCOM,".com");
		
	file = fopen( "AGI", "wb" );
	fwrite( agi, 1, 39424, file );
	fclose(file);
	
	file = fopen( "AGIDATA.OVL", "wb" );
	fwrite( agidata, 1, 8192, file );
	fclose(file);
	
	file = fopen( "CGA_GRAF.OVL", "wb" );
	fwrite( cgagraf, 1, 1024, file );
	fclose(file);
	
	file = fopen( "EGA_GRAF.OVL", "wb" );
	fwrite( egagraf, 1, 1024, file );
	fclose(file);
	
	file = fopen( "HGC_FONT", "wb" );
	fwrite( hgcfont, 1, 3072, file );
	fclose(file);
	
	file = fopen( "HGC_GRAF.OVL", "wb" );
	fwrite( hgcgraf, 1, 1536, file );
	fclose(file);
	
	file = fopen( "HGC_OBJS.OVL", "wb" );
	fwrite( hgcobjs, 1, 1024, file );
	fclose(file);
	
	file = fopen( "IBM_OBJS.OVL", "wb" );
	fwrite( ibmobjs, 1, 512, file );
	fclose(file);
	
	file = fopen( "JR_GRAF.OVL", "wb" );
	fwrite( jrgraf, 1, 512, file );
	fclose(file);
	
	file = fopen( "VG_GRAF.OVL", "wb" );
	fwrite( vgagraf, 1, 512, file );
	fclose(file);
	
	file = fopen(gameCOM, "wb" );
	fwrite( sierracom, 1, 3115, file );
	fclose(file);
	printf("Use %s to run your game.\n", gameCOM);
}

void patchAgiWithID(char* id)
{
	int offset = 23716;
	int i;
	for(i = 0; id[i] != '\0'; i++) 
	{
		agi[offset+i] = id[i];
	}
	/* write last null byte to end the string*/
	agi[offset+i] = 0;
}
 
void checkInterp(char* foundInterp)
{
	char * interpid = "2.917";
	int c,temp;
	
	if(foundInterp[0]!='2')
	{
		printf("This is not an AGI 2.x game, and cannot be updated. \n");
		exit(1);
	}
	
	if(foundInterp[2]=='0' || foundInterp[2]=='2')
	{
		printf("Your game is designed for an interpreter which is too old to be upgraded. \n");
		exit(1);
	}
	
	if(foundInterp[2]=='9' && foundInterp[3]=='3' && foundInterp[4]=='6')
	{
		printf("Your game has a newer interpreter than 2.917. Quitting... \n");
		exit(1);
	}
	
	if((strcmp(interpid,foundInterp)) == 0)
	{
		printf("Do you still want to replace the interpreter files? (y/n):");
		c = getchar();
		while ((temp = getchar()) != '\n' && temp != EOF) { }
		if(c!='y')
		{
			exit(1);
		}
	}
}	
 
int main(int argc, char **argv)
{
	char * gameID;
	int i;
    FILE *fp;
	char* foundInterp;
	int c;
	char * buffer;
	int count;
	
	/* printf("%d %s \n",argc,argv[0]); */

	foundInterp = getInterpVer();
	printf("Current game files are using interpreter %s \n",foundInterp);
	checkInterp(foundInterp);

	c=0;
	count=0;
	do
	{
		printf("\nPlease pick which game you are upgrading the interpreter of: \n");
		printf("[1] Leisure Suite Larry 1 \n");
		printf("[2] KQ1 \n");
		printf("[3] KQ2 \n");
		printf("[4] KQ3 \n");
		printf("[5] Mixed Up Mother Goose \n");
		printf("[6] PQ1 \n");
		printf("[7] SQ1 \n");
		printf("[8] SQ2 \n");
		printf("[9] Black Caldron \n");
		printf("[10] Donald Ducks Playground \n");
		printf("[11] Something not listed that I know the game ID for \n");
		printf("[q] Quit! \n");
		printf(":");
		
		buffer = (char *)malloc(sizeof(char)*256);
		while ((c = getchar()) != '\n' && c != EOF && count < 255)
		{
			buffer[count] = c;
			count++;
		}
		buffer[count]=0;
		
		/*printf("%s", buffer);*/
		
		if(((strcmp(buffer,"1")) == 0) || ((strcmp(buffer,"2")) == 0) || ((strcmp(buffer,"3")) == 0) || ((strcmp(buffer,"4")) == 0) || ((strcmp(buffer,"5")) == 0) || ((strcmp(buffer,"6")) == 0) || ((strcmp(buffer,"7")) == 0) || ((strcmp(buffer,"8")) == 0) || ((strcmp(buffer,"9")) == 0) || ((strcmp(buffer,"10")) == 0) || ((strcmp(buffer,"11")) == 0))
			{
				c = 0;
				if((strcmp(buffer,"1")) == 0)
				{
					gameID = "LLLLL";
				}
				if((strcmp(buffer,"2")) == 0)
				{
					gameID = "KQ1";
				}
				if((strcmp(buffer,"3")) == 0)
				{
					gameID = "KQ2";
				}
				if((strcmp(buffer,"4")) == 0)
				{
					gameID = "KQ3";
				}
				if((strcmp(buffer,"5")) == 0)
				{
					gameID = "MG";
				}
				if((strcmp(buffer,"6")) == 0)
				{
					gameID = "PQ";
				}
				if((strcmp(buffer,"7")) == 0)
				{
					gameID = "SQ";
				}
				if((strcmp(buffer,"8")) == 0)
				{
					gameID = "SQ2";
				}
				if((strcmp(buffer,"9")) == 0)
				{
					gameID = "BC";
				}
				if((strcmp(buffer,"10")) == 0)
				{
					gameID = "DUCK";
				}
				if((strcmp(buffer,"11")) == 0)
				{
					free(buffer);
					buffer = (char *)malloc(sizeof(char)*6);
					count=0;
					c=0;
					
					printf("Please enter a gameID (5 characters or less):");
					while ((c = getchar()) != '\n' && c != EOF && count < 5)
					{
						buffer[count] = c;
						count++;
					}
					buffer[count]=0;
					gameID = (char *)malloc(sizeof(char)*6);
					strcpy(gameID,buffer);
					free(buffer);
					c=0;
				}
			}
		else if((strcmp(buffer,"q") == 0))
		{
			exit(1);
		}
		else
		{
			c=-1;
		}
		
		free(buffer);
		count=0;
	} while(c!=0);
	
	printf("\nUpgrading %s \n", gameID);
	
	printf("Backing up game files and deleting \n");
	removeGameFiles();
	printf("Writing new game files \n");
	patchAgiWithID(gameID);
	writeNewGameFiles(gameID);

 
    return 0;
}